from wpilib.command import Command


def checkIfCanceled(self):
    if self.forceCancel:
        self.forceCancel = False
        return True

    return self._isFinished()


class CancelCommand(Command):
    """
    When this command is run, it cancels the command it was passed.
    """

    def __init__(self, command: Command):
        """
        :param command: The command to cancel.
        """
        super().__init__("Cancel %s" % command)

        self.command = command

    def initialize(self):
        if self.command.isRunning():
            self.command.cancel()

    def isFinished(self):
        return not self.command.isRunning()



from wpilib import TimedRobot
from wpilib.command import Scheduler


class CommandBasedRobot(TimedRobot):
    """
    The base class for a Command-Based Robot. To use, instantiate commands and
    trigger them.
    """

    def startCompetition(self):
        """Initalizes the scheduler before starting robotInit()"""

        self.scheduler = Scheduler.getInstance()
        super().startCompetition()

    def commandPeriodic(self):
        """
        Run the scheduler regularly. If an error occurs during a competition,
        prevent it from crashing the program.
        """

        try:
            self.scheduler.run()
        except Exception as error:
            if not self.ds.isFMSAttached():
                raise

            """Just to be safe, stop all running commands."""
            self.scheduler.removeAll()

            self.handleCrash(error)

    autonomousPeriodic = commandPeriodic
    teleopPeriodic = commandPeriodic
    disabledPeriodic = commandPeriodic
    # testPeriodic deliberately omitted

    def handleCrash(self, error):
        """
        Called if an exception is raised in the Scheduler during a competition.
        Writes an error message to the driver station by default. If you want
        more complex behavior, override this method in your robot class.
        """

        self.ds.reportError(str(error), printTrace=True)



from . import _init_commands_v1

# autogenerated by 'robotpy-build create-imports wpilib.command wpilib.command._commands_v1'
from ._commands_v1 import (
    Button,
    ButtonScheduler,
    CancelButtonScheduler,
    Command,
    CommandGroup,
    CommandGroupEntry,
    ConditionalCommand,
    HeldButtonScheduler,
    InstantCommand,
    InternalButton,
    JoystickButton,
    NetworkButton,
    PIDCommand,
    PIDSubsystem,
    POVButton,
    PressedButtonScheduler,
    PrintCommand,
    ReleasedButtonScheduler,
    Scheduler,
    StartCommand,
    Subsystem,
    TimedCommand,
    ToggleButtonScheduler,
    Trigger,
    WaitCommand,
    WaitForChildren,
    WaitUntilCommand,
)

__all__ = [
    "Button",
    "ButtonScheduler",
    "CancelButtonScheduler",
    "Command",
    "CommandGroup",
    "CommandGroupEntry",
    "ConditionalCommand",
    "HeldButtonScheduler",
    "InstantCommand",
    "InternalButton",
    "JoystickButton",
    "NetworkButton",
    "PIDCommand",
    "PIDSubsystem",
    "POVButton",
    "PressedButtonScheduler",
    "PrintCommand",
    "ReleasedButtonScheduler",
    "Scheduler",
    "StartCommand",
    "Subsystem",
    "TimedCommand",
    "ToggleButtonScheduler",
    "Trigger",
    "WaitCommand",
    "WaitForChildren",
    "WaitUntilCommand",
]

del _init_commands_v1

try:
    from .version import version as __version__
except ImportError:
    __version__ = "master"



# import pytest

# from wpilib.command import Scheduler, InstantCommand

import commandbased.flowcontrol as fc


# def run_command(cmd):
#     cmd.start()

#     scheduler = Scheduler.getInstance()
#     for x in range(0, 2000):
#         scheduler.run()
#         if not cmd.isRunning():
#             return True

#     return False


# class CounterCommand(InstantCommand):
#     def __init__(self):
#         super().__init__("Counter")
#         self.counter = 0

#     def initialize(self):
#         self.counter += 1


# def test_if():
#     cmd1 = CounterCommand()
#     cmd2 = CounterCommand()

#     class IfCommand(fc.CommandFlow):
#         def __init__(self):
#             nonlocal cmd1, cmd2

#             super().__init__("Test IF")
#             self.setRunWhenDisabled(True)

#             @fc.IF(lambda: False)
#             def increment_cmd1(self):
#                 self.addSequential(cmd1)

#             @fc.IF(lambda: True)
#             def increment_cmd2(self):
#                 self.addSequential(cmd2)

#     cmd = IfCommand()

#     assert run_command(cmd)
#     assert cmd1.counter == 0
#     assert cmd2.counter == 1


# def test_elif():
#     cmd1 = CounterCommand()
#     cmd2 = CounterCommand()
#     cmd3 = CounterCommand()

#     class ElifCommand(fc.CommandFlow):
#         def __init__(self):
#             nonlocal cmd1, cmd2, cmd3

#             super().__init__("Test ELIF")
#             self.setRunWhenDisabled(True)

#             @fc.IF(lambda: False)
#             def increment_cmd1(self):
#                 self.addSequential(cmd1)

#             @fc.ELIF(lambda: True)
#             def increment_cmd2(self):
#                 self.addSequential(cmd2)

#             @fc.ELIF(lambda: True)
#             def increment_cmd3(self):
#                 self.addSequential(cmd3)

#     cmd = ElifCommand()

#     assert run_command(cmd)
#     assert cmd1.counter == 0
#     assert cmd2.counter == 1
#     assert cmd3.counter == 0


# def test_else():
#     cmd1 = CounterCommand()
#     cmd2 = CounterCommand()
#     cmd3 = CounterCommand()

#     class ElseCommand(fc.CommandFlow):
#         def __init__(self):
#             nonlocal cmd1, cmd2, cmd3

#             super().__init__("Test ELSE")
#             self.setRunWhenDisabled(True)

#             @fc.IF(lambda: False)
#             def increment_cmd1(self):
#                 self.addSequential(cmd1)

#             @fc.ELIF(lambda: False)
#             def increment_cmd2(self):
#                 self.addSequential(cmd2)

#             @fc.ELSE
#             def increment_cmd3(self):
#                 self.addSequential(cmd3)

#     cmd = ElseCommand()

#     assert run_command(cmd)
#     assert cmd1.counter == 0
#     assert cmd2.counter == 0
#     assert cmd3.counter == 1


# # def test_while():
# #     cmd1 = CounterCommand()

# #     class WhileCommand(fc.CommandFlow):
# #         def __init__(self):
# #             nonlocal cmd1

# #             super().__init__("Test WHILE")
# #             self.setRunWhenDisabled(True)

# #             x = 5

# #             def do_loop():
# #                 nonlocal x

# #                 x -= 1

# #                 return x > 0

# #             @fc.WHILE(do_loop)
# #             def count_to_four(self):
# #                 self.addSequential(cmd1)

# #     cmd = WhileCommand()

# #     assert run_command(cmd)
# #     assert cmd1.counter == 4


# # def test_break():
# #     cmd1 = CounterCommand()
# #     cmd2 = CounterCommand()

# #     class BreakCommand(fc.CommandFlow):
# #         def __init__(self):
# #             nonlocal cmd1, cmd2

# #             super().__init__("Test BREAK")
# #             self.setRunWhenDisabled(True)

# #             x = 5

# #             def do_loop():
# #                 nonlocal x

# #                 x -= 1

# #                 return x > 0

# #             @fc.WHILE(do_loop)
# #             def count_to_four(self):
# #                 self.addSequential(cmd1)

# #                 @fc.IF(lambda: x == 3)
# #                 def breakLoop(self):
# #                     fc.BREAK()

# #             self.addSequential(cmd2)

# #     cmd = BreakCommand()

# #     assert run_command(cmd)
# #     assert cmd1.counter == 2
# #     assert cmd2.counter == 1


# def test_return():
#     cmd1 = CounterCommand()
#     cmd2 = CounterCommand()
#     cmd3 = CounterCommand()
#     cmd4 = CounterCommand()
#     cmd5 = CounterCommand()
#     cmd6 = CounterCommand()

#     class ReturnCommand(fc.CommandFlow):
#         def __init__(self):
#             nonlocal cmd1, cmd2, cmd3, cmd4, cmd5, cmd6

#             super().__init__("Test RETURN")
#             self.setRunWhenDisabled(True)

#             self.addSequential(cmd1)

#             @fc.IF(lambda: True)
#             def increment_cmd2(self):
#                 self.addSequential(cmd2)

#                 @fc.IF(lambda: True)
#                 def increment_cmd3(self):
#                     self.addSequential(cmd3)

#                     @fc.IF(lambda: True)
#                     def do_return(self):
#                         fc.RETURN()

#                     self.addSequential(cmd4)

#                 self.addSequential(cmd5)

#             self.addSequential(cmd6)

#     cmd = ReturnCommand()

#     assert run_command(cmd)
#     assert cmd1.counter == 1
#     assert cmd2.counter == 1
#     assert cmd3.counter == 1
#     assert cmd4.counter == 0
#     assert cmd5.counter == 0
#     assert cmd6.counter == 0


# # def test_break2():
# #     cmd1 = CounterCommand()
# #     cmd2 = CounterCommand()
# #     cmd3 = CounterCommand()

# #     class BreakCommand(fc.CommandFlow):
# #         def __init__(self):
# #             nonlocal cmd1, cmd2, cmd3

# #             super().__init__("Test BREAK 2")
# #             self.setRunWhenDisabled(True)

# #             x = 5

# #             def do_loop():
# #                 nonlocal x

# #                 x -= 1

# #                 if x == 0:
# #                     x = 10
# #                     return False

# #                 return True

# #             @fc.WHILE(lambda: True)
# #             def loop_forever(self):
# #                 self.addSequential(cmd1)

# #                 @fc.WHILE(do_loop)
# #                 def count_to_four(self):
# #                     self.addSequential(cmd2)

# #                     @fc.IF(lambda: x == 5)
# #                     def breakLoop(self):
# #                         fc.BREAK(2)

# #             self.addSequential(cmd3)

# #     cmd = BreakCommand()

# #     assert run_command(cmd)
# #     assert cmd1.counter == 2
# #     assert cmd2.counter == 9
# #     assert cmd3.counter == 1


# def test_contained_return():
#     cmd1 = CounterCommand()
#     cmd2 = CounterCommand()
#     cmd3 = CounterCommand()
#     cmd4 = CounterCommand()

#     class ReturnCommand(fc.CommandFlow):
#         def __init__(self):
#             nonlocal cmd2, cmd3

#             super().__init__("Test RETURN")
#             self.setRunWhenDisabled(True)

#             self.addSequential(cmd2)
#             fc.RETURN()
#             self.addSequential(cmd3)

#     class ContainerCommand(fc.CommandFlow):
#         def __init__(self):
#             nonlocal cmd1, cmd4

#             super().__init__("Test Contained RETURN")
#             self.setRunWhenDisabled(True)

#             self.addSequential(cmd1)
#             self.addSequential(ReturnCommand())
#             self.addSequential(cmd4)

#     cmd = ContainerCommand()

#     assert run_command(cmd)
#     assert cmd1.counter == 1
#     assert cmd2.counter == 1
#     assert cmd3.counter == 0
#     assert cmd4.counter == 1


# def test_safety():
#     class ElifBeforeIfCommand(fc.CommandFlow):
#         def __init__(self):
#             super().__init__("Test ELIF before IF")

#             @fc.ELIF(lambda: True)
#             def do_nothing(self):
#                 self.addSequential(InstantCommand())

#     with pytest.raises(ValueError):
#         cmd = ElifBeforeIfCommand()

#     class ElseBeforeIfCommand(fc.CommandFlow):
#         def __init__(self):
#             super().__init__("Test ELSE before IF")

#             @fc.ELSE
#             def do_nothing(self):
#                 self.addSequential(InstantCommand())

#     with pytest.raises(ValueError):
#         cmd = ElseBeforeIfCommand()

#     # class BreakTooBigCommand(fc.CommandFlow):
#     #     def __init__(self):
#     #         super().__init__("Test BREAK too large")
#     #         self.setRunWhenDisabled(True)

#     #         @fc.WHILE(lambda: True)
#     #         def do_loop(self):
#     #             fc.BREAK(2)

#     # cmd = BreakTooBigCommand()

#     # with pytest.raises(ValueError):
#     #     run_command(cmd)

#     # class BreakWithoutLoop(fc.CommandFlow):
#     #     def __init__(self):
#     #         super().__init__("Test BREAK outside loop")
#     #         self.setRunWhenDisabled(True)

#     #         fc.BREAK()

#     # with pytest.raises(ValueError):
#     #     cmd = BreakWithoutLoop()

#     # class BreakAfterLoop(fc.CommandFlow):
#     #     def __init__(self):
#     #         super().__init__("Test BREAK after loop")
#     #         self.setRunWhenDisabled(True)

#     #         x = 5

#     #         def do_loop():
#     #             nonlocal x

#     #             x -= 1

#     #             return x > 0

#     #         @fc.WHILE(do_loop)
#     #         def count_to_four(self):
#     #             self.addSequential(InstantCommand())

#     #         @fc.IF(lambda: True)
#     #         def late_break(self):
#     #             fc.BREAK()

#     # with pytest.raises(ValueError):
#     #     cmd = BreakAfterLoop()

#     with pytest.raises(ValueError):
#         fc.RETURN()



"""
These functions can be used to make programming CommandGroups much more
intuitive. For more information, check each method's docstring.
"""

from wpilib.command import Command, CommandGroup, ConditionalCommand
from commandbased.cancelcommand import CancelCommand

import inspect

__all__ = ["IF", "ELIF", "ELSE", "WHILE", "RETURN", "BREAK", "CommandFlow"]


class ConditionalFlow(ConditionalCommand):
    def __init__(self, name, onTrue, onFalse, condition):
        ConditionalCommand.__init__(self, name, onTrue, onFalse)

        self.flowCondition = condition

    def _condition(self):
        return self.flowCondition()


class CommandFlow(CommandGroup):
    def __init__(self, name):
        CommandGroup.__init__(self, name)

        callingFlow = _getCommandFlow()
        self._source = getattr(callingFlow, "_source", self)
        self._ifStack = None

    def _popIfStack(self):
        """
        We buffer conditionals until the last moment so we don't have trouble with
        Commands being locked when they're added to a CommandGroup.
        """
        if self._ifStack:
            top = self._ifStack.pop(0)
            cmd = None
            for x in reversed(self._ifStack):
                if x[0]:
                    cmd = ConditionalFlow("flowcontrolELIF", x[1], cmd, x[0])

                else:
                    cmd = x[1]

            cmd = ConditionalFlow("flowcontrolIF", top[1], cmd, top[0])

            self._ifStack = None
            self.addSequential(cmd)

    # These _hook methods ensure we always add our buffered conditions
    def addSequential(self, cmd, timeout=None):
        self._popIfStack()
        if timeout is None:
            CommandGroup.addSequential(self, cmd)
        else:
            CommandGroup.addSequential(self, cmd, timeout)

    def addParallel(self, cmd, timeout=None):
        self._popIfStack()
        if timeout is None:
            CommandGroup.addParallel(self, cmd)
        else:
            CommandGroup.addParallel(self, cmd, timeout)

    def start(self):
        self._popIfStack()
        CommandGroup.start(self)

    def setParent(self, parent):
        self._popIfStack()
        CommandGroup.setParent(self, parent)


class CommandFlowWhile(CommandFlow):
    def __init__(self, name, condition):
        super().__init__(name)

        self.whileCondition = condition

    def isFinished(self):
        if CommandGroup.isFinished(self):
            if not self.whileCondition():
                return True

            self.start()

        return False


def _getCommandFlow():
    """
    Does some rather ugly stack inspection to find out which CommandGroup the
    calling function was triggered from.
    """
    # https://stackoverflow.com/a/14694234
    stack = inspect.stack()
    frame = stack[2].frame
    while "self" not in frame.f_locals:
        frame = frame.f_back
        if frame is None:
            raise ValueError("Could not find calling class for %s" % stack[1].function)

    cg = frame.f_locals["self"]
    if not isinstance(cg, CommandFlow):
        raise ValueError(
            "%s may not be used outside of a CommandFlow" % stack[1].function
        )

    return cg


def _buildCommandFlow(func, parent):
    """Turns the given function into a full CommandGroup."""
    cg = CommandFlow(func.__name__)
    func(cg)

    return cg


def IF(condition):
    """
    Use as a decorator for a function. That function will be placed into a
    CommandGroup and run inside a ConditionalCommand with the given condition.
    The decorated function must accept one positional argument that will be used
    as its 'self'.
    """

    def flowcontrolIF(func):
        parent = _getCommandFlow()
        cg = _buildCommandFlow(func, parent)

        try:
            parent._popIfStack()
        except AttributeError:
            pass

        parent._ifStack = [(condition, cg)]

    return flowcontrolIF


def ELIF(condition):
    """
    Use as a decorator for a function. That function will be placed into a
    CommandGroup which will be triggered by a ConditionalCommand that uses the
    passed condition. That ConditionalCommand will then be added as the onFalse
    for the ConditionalCommand created by a previous IF or ELIF.
    """

    def flowcontrolELIF(func):
        parent = _getCommandFlow()
        cg = _buildCommandFlow(func, parent)

        try:
            parent._ifStack.append((condition, cg))
        except AttributeError:
            raise ValueError("Cannot use ELIF without IF")

    return flowcontrolELIF


def ELSE(func):
    """
    Use as a decorator for a function. That function will be placed into a
    CommandGroup which will be added as the onFalse for the ConditionalCommand
    created by a previous IF or ELIF.
    """

    parent = _getCommandFlow()
    cg = _buildCommandFlow(func, parent)

    try:
        parent._ifStack.append((None, cg))
    except AttributeError:
        raise ValueError("Cannot use ELSE without IF")

    parent._popIfStack()


def WHILE(condition):
    """
    Use as a decorator for a function. That function will be placed into a
    CommandGroup, which will be added to a ConditionalCommand. It will be
    modified to restart itself automatically.
    """
    raise NotImplementedError("WHILE does not yet work with Commands v1")


def RETURN():
    """
    Calling this function will end the source CommandGroup immediately.
    """

    parent = _getCommandFlow()
    parent.addSequential(CancelCommand(parent._source))


def BREAK(steps=1):
    """
    Calling this function will end the loop that contains it. Pass an integer to
    break out of that number of nested loops.
    """
    raise ValueError("Cannot BREAK outside of a loop")
